import time
import librosa
try:
    from .cvt import *
except ImportError:
    from cvt import *

# 记录程序开始执行的时间
START_TIME = time.time()

def show_elapsed_time():
    """显示从程序开始执行到现在的时间差"""
    elapsed = time.time() - START_TIME
    minutes = int(elapsed // 60)
    seconds = elapsed % 60
    milliseconds = int((elapsed % 1) * 1000)
    return f"{minutes:02d}:{int(seconds):02d}:{milliseconds:03d}"


def read_audio(audio_path, tar_sr=None):
    y, sr = librosa.load(audio_path, mono=False, sr=tar_sr)
    # print(y.shape, sr)
    if y.ndim >= 2:
        y = y[0]
        
    return y, sr


def bandpass_filter(data, lowcut, highcut, fs, order=4):
    nyquist = 0.5 * fs
    low = lowcut / nyquist
    high = highcut / nyquist
    if low == 0:
        b, a = butter(order, high, btype='low', output="ba")
        filtered_data = filtfilt(b, a, data)
    else:
        try:
            b, a = butter(order, [low, high], btype='bandpass', output="ba")
            filtered_data = filtfilt(b, a, data)
        except ValueError:  # 如果设置的最高频率大于了可接受的范围
            b, a = butter(order, low, btype='high', output="ba")
            filtered_data = filtfilt(b, a, data)
    return filtered_data



def extract_cvt(character, lang):
    if lang == 'zh':
        return extract_cvt_zh(character)
    elif lang == "yue":
        return extract_cvt_yue(character)
    else:
        raise ValueError(f"{show_elapsed_time()}Language {lang} not yet supported.")


if __name__ == '__main__':
    print(extract_cvt('我', 'zh'))


try:
    from .utils import *
    from .cvt import *
except ImportError:
    from utils import *
    from cvt import *

import librosa
import numpy as np
from scipy.signal import find_peaks
from textgrid import TextGrid
import parselmouth



def get_expected_num(audio_path):
    tg = TextGrid.fromFile(audio_path.replace(".wav", "_whisper.TextGrid"))
    tier = tg.tiers[0]
    words = []
    for interval in tier.intervals:
        for c in interval.mark.split():
            words.append(c)

    count = 0
    for word in words:
        c, v, t = extract_cvt(word, lang="zh")[0]
        if c in ["z", "zh", "s", "c", "ch", "sh", "x", "j"]:
            count += 1
    return count


def get_spectral_peak_interval(audio_path, verbose=False):
    """
    绘制音频的频谱质心曲线
    
    参数:
    audio_path (str): 音频文件的路径
    """

    vad_path = audio_path.replace(".wav", "_VAD.TextGrid")
    whisper_path = audio_path.replace(".wav", "_whisper.TextGrid")
    # 加载音频文件
    y, sr = read_audio(audio_path)
    # print(y.shape, sr)
    
    
    # 计算频谱质心
    # 计算每个帧的频谱
    stft = librosa.stft(y)#, win_length=win_length, hop_length=hop_length)
    # 计算每个帧的幅度谱
    magnitude = np.abs(stft)
    # 找到每个帧中频谱幅度最大的索引
    max_magnitude_indices = np.argmax(magnitude, axis=0)
    # 将索引转换为对应的频率
    spectral_peaks = librosa.fft_frequencies(sr=sr)[max_magnitude_indices]
    
    # 计算对应的时间轴
    time = librosa.times_like(spectral_peaks, sr=sr)#, hop_length=hop_length)



    # 筛选出 start_time 到 end_time 之间的数据
    # mask = (time >= start_time) & (time <= end_time)
    # time = time[mask]
    # spectral_peaks = spectral_peaks[mask]
    
    # 定义平均池化的窗口大小
    window_size = int(0.0001 * sr)
    # 创建一个归一化的卷积核
    kernel = np.ones(window_size) / window_size
    # 使用 np.convolve 进行平均池化
    spectral_peaks = np.convolve(spectral_peaks, kernel, mode='same')


    # 方法一：获取baseline
    # basline = get_baseline_freq_peak(audio_path)  
    # 方法二：根据whisper的标注获取预期的音节数

    tg = TextGrid.fromFile(audio_path.replace(".wav", "_whisper.TextGrid"))
    tier = tg.tiers[0]
    words = []
    for interval in tier.intervals:
        for c in interval.mark.split():
            words.append(c)

    count = 0
    for word in words:
        c, v, t = extract_cvt(word, lang="zh")[0]
        if c in ["z", "zh", "s", "c", "ch", "sh", "x", "j"]:
            count += 1
    if verbose:
        print(f"[{show_elapsed_time()}] Expected number of syllables: {count}")
    expected_num = count




    ifMeetNum = None
    # 从spectral_peaks的最大值往0遍历，找到最后一个符合expected_num的baseline
    max_peak = np.max(spectral_peaks)
    step = max_peak / 100  # 设置遍历步长

    # for baseline in np.arange(max_peak, 0, -step):
    for baseline in np.arange(0, max_peak, step):
        # 找到所有大于基线的点
        above_baseline_mask = spectral_peaks > baseline
        # 找到所有大于基线的点的连续区间
        continuous_count = 0
        in_interval = False

        cand_timestamps = []
        for idx, is_above in enumerate(above_baseline_mask):
            if is_above and not in_interval:
                # continuous_count += 1
                in_interval = True
                cand_timestamps.append([time[idx]])
            elif not is_above and in_interval:
                cand_timestamps[-1].append(time[idx])
                in_interval = False
            # elif not in_interval:
                # below_frm_num += 1
        # if len(cand_timestamps) >= 1:
        #     continuous_count += 1
        # # 检查从false到true之间的距离是否够大，若够大，才考虑切换
        # for idx in range(1, len(cand_timestamps)):

        #     if cand_timestamps[idx][0] - cand_timestamps[idx-1][1] > 0.005:
        #         continuous_count += 1
        if len(cand_timestamps) == 1 and len(cand_timestamps[0]) == 1:
            cand_timestamps[-1].append(time[-1])
        continuous_count = len(cand_timestamps)

        if verbose:
            print(cand_timestamps, continuous_count, expected_num)
        # 检查最后一个区间是否以大于基线的值结束
        if len(above_baseline_mask) > 0 and above_baseline_mask[-1] and in_interval:
            pass
        # !!! 需要添加：相邻两个peak之间不能太近
        # 若连续区间数量等于预期数量，则使用当前基线
        if ifMeetNum is None:
            if continuous_count == expected_num:
                break
                ifMeetNum = True
    
        elif ifMeetNum:
            if continuous_count != expected_num:
                # print(continuous_count, expected_num, "-")
                ifMeetNum = False
                break
        
        # elif not ifMeetNum:
        #     print(continuous_count, expected_num, "!")
        #     print()
        #     break
    # print(baseline)
    # baseline -= step
    if verbose:
        print(f"Baseline: {baseline:.2f}, Continuous Count: {continuous_count}, Expected Num: {expected_num}, ifMeetNum: {ifMeetNum}")

    # 找到所有大于基线的点
    above_baseline_mask = spectral_peaks > baseline

    # 初始化存储连续区间的列表
    continuous_intervals = []
    start_index = None

    # 遍历掩码，找到连续大于基线的区间
    for idx, is_above in enumerate(above_baseline_mask):
        if is_above and start_index is None:
            start_index = idx
        elif not is_above and start_index is not None:
            continuous_intervals.append((start_index, idx))
            start_index = None

    # 处理最后一个区间
    if start_index is not None:
        continuous_intervals.append((start_index, len(above_baseline_mask)))

    # 存储每个区间最大值的索引
    # max_indices = []
    interval_indices = []
    for idx, (start, end) in enumerate(continuous_intervals):
        # interval_data = spectral_peaks[start:end]
        # max_index_in_interval = np.argmax(interval_data)
        # max_indices.append(start + max_index_in_interval)

        # 找到start和end之间的time上的所有波峰
        interval_spectral_peaks = spectral_peaks[start:end]
        peak_indices, _ = find_peaks(interval_spectral_peaks)
        peak_indices = list(sorted(peak_indices))
        if peak_indices:
            start = start + peak_indices[0]
            end = start + peak_indices[-1]

        dur = (time[end] - time[start]) * 0.  # 也可以去找最近的波峰!!!

        left = time[start] + dur
        right = time[end] - dur

        if idx == 0:
            interval_indices.append([0.0, left])
            interval_indices.append([right])
        elif idx == len(continuous_intervals) - 1:
            interval_indices[-1].append(left)
            interval_indices.append([right, len(y)/sr])
        else:
            interval_indices[-1].append(left)
            interval_indices.append([right])
    # print(interval_indices)
    # interval_indices = [i for i in interval_indices if min_pause < i[1] - i[0]]
    # max_indices = np.array(max_indices)
    # print(interval_indices)
    

    if verbose:
        import matplotlib.pyplot as plt

        # 创建图形
        plt.figure(figsize=(10, 4))
        
        # 绘制频谱质心曲线
        plt.plot(time, spectral_peaks, color='r')
        plt.title('Audio Spectral Centroid Curve')
        plt.xlabel('Time (s)')
        plt.ylabel('Spectral Centroid (Hz)')
        for start, end in interval_indices:
            plt.axvline(x=start, color='b', linestyle='--')
            plt.axvline(x=end, color='g', linestyle='--')
        plt.grid(True)
        
        # 显示图形
        plt.tight_layout()
        plt.show()
    # print(time[np.argmax(spectral_peaks)])

    return interval_indices


def find_internsity_valley(audio_path, start_time, end_time, verbose=False):
    sound = parselmouth.Sound(audio_path)

    # 计算整个音频的强度对象
    intensity = sound.to_intensity(time_step=0.02)  # 时间步长0.01秒（可调整）

    intensity_points = np.array(intensity.as_array()).flatten()
    time_points = np.array(intensity.xs())
    # 筛选出 current_interval.minTime 和 next_interval.maxTime 之间的时间点和对应的强度值
    mask = (time_points >= start_time) & (time_points <= end_time)
    time_points = time_points[mask]
    intensity_points = intensity_points[mask]

    # if verbose:
    #     import matplotlib.pyplot as plt

    #     # 创建图形
    #     plt.figure(figsize=(10, 4))
        
    #     # 绘制功率曲线
    #     plt.plot(time_points, intensity_points, alpha=0.3)
    #     plt.title('Audio Power Curve')
    #     plt.xlabel('Time (s)')
    #     plt.ylabel('Power')
    #     plt.grid(True)
    # 找到强度曲线的波谷索引
    intensity_valley_indices = find_peaks(-intensity_points)[0]

    # midpoint = (start_time + end_time) / 2
    # 按照距离 midpoint 的绝对距离对波谷索引排序
    # intensity_valley_indices = sorted(intensity_valley_indices, key=lambda idx: abs(time_points[idx] - midpoint))
    intensity_valley_indices = sorted(intensity_valley_indices, key=lambda idx: intensity_points[idx])
    # 获取波谷对应的时间点
    valley_times = time_points[intensity_valley_indices]
    try:
        min_valley_time = valley_times[0]
    except IndexError:
        min_valley_time = time_points[np.argmin(intensity_points)]
    
    # if verbose:
    #     plt.axvline(x=min_valley_time, color='b', linestyle='--')
    #     plt.show()

    return min_valley_time




def find_word_boundary(audio_path, tar_sr=10000, min_pause=0.1, verbose=False):
    """
    绘制整段音频的功率曲线
    
    参数:
    audio_path (str): 音频文件的路径
    """
    shifted_peaks_indices = get_spectral_peak_interval(audio_path, verbose=verbose)


    tg_path = audio_path.replace(".wav", "_whisper.TextGrid")
    # print(shifted_peaks_indices)

    # 加载音频文件
    y, sr = read_audio(audio_path, tar_sr=tar_sr)


    y = np.gradient(y)
    # y = np.gradient(np.gradient(y))

    # y = bandpass_filter(y, 50, sr, sr, order=4)

    # 使用librosa计算音频的均方根能量(rms)
    rms = librosa.feature.rms(y=y, frame_length=128, hop_length=32, center=True)[0]

    # 计算对应的时间轴
    time = librosa.times_like(rms, sr=sr, hop_length=32)
    
    if verbose:
        import matplotlib.pyplot as plt

        # 创建图形
        plt.figure(figsize=(10, 4))
        
        # 绘制功率曲线
        plt.plot(time, rms, alpha=0.3)
        plt.title('Audio Power Curve')
        plt.xlabel('Time (s)')
        plt.ylabel('Power')
        plt.grid(True)

        vertical_line = [.688, .80, .88, 1.16, 1.25, 1.55, 1.75, 1.84, 1.94, 2.224, 2.49,
                        3.51, 3.75, 3.97, 4.10, 4.29, 4.46, 4.58, 4.72, 4.958]
        for v in vertical_line:
            plt.axvline(x=v, color='r', linestyle='--')

    # 找到波谷的索引
    valley_indices = find_peaks(-rms, width=(1, None), distance=10)[0]

    if verbose:
        plt.scatter(time[valley_indices], rms[valley_indices], color='orange', label='Valley')


    tg = TextGrid()
    tg.read(tg_path)
    intervals = [interval for interval in tg.tiers[0] if interval.mark != ""]

    for idx, interval in enumerate(intervals):
        if idx == len(intervals) - 1:
            break

            
        current_interval = intervals[idx]
        next_interval = intervals[idx+1]

        current_con, current_vow, current_tone = extract_cvt_zh(current_interval.mark)[0]
        next_con, next_vow, next_tone = extract_cvt_zh(next_interval.mark)[0]

        if current_interval.maxTime != next_interval.minTime:
            continue
            

        cand_valleys = [t for t in time[valley_indices] if current_interval.minTime + 0.05 < t < next_interval.maxTime]
        cand_valleys_rms = [rms[np.where(time == t)[0][0]] for t in cand_valleys]


        # 将候选波谷时间转换为 numpy 数组以便后续操作
        valid_valleys = np.array(cand_valleys)
        valid_valleys_rms = np.array(cand_valleys_rms)
        

        isNextConFlag = next_con in ["z", "zh", "s", "c", "ch", "sh", "x", "j"]
        isCurrentConFlag = current_con in ["z", "zh", "s", "c", "ch", "sh", "x", "j"]

        sorted_indices = np.argsort(valid_valleys_rms)

        best_dur = 0
        left_boundary = None
        right_boundary = None

        for start, end in shifted_peaks_indices:
            if isCurrentConFlag and isNextConFlag:
                if (current_interval.minTime <= start <= next_interval.maxTime) and (current_interval.minTime <= end <= next_interval.maxTime):
                    dur = min(end, next_interval.maxTime) - max(start, current_interval.minTime)

                    if dur > best_dur:
                        best_dur = dur

                        left_boundary = max(start, current_interval.minTime)
                        right_boundary = min(end, next_interval.maxTime)
            
            elif isCurrentConFlag and not isNextConFlag:
                if current_interval.minTime <= start <= next_interval.maxTime:
                
                    dur = next_interval.maxTime - max(start, current_interval.minTime)

                    if dur > best_dur:
                        best_dur = dur

                        left_boundary = max(start, current_interval.minTime)
                        right_boundary = next_interval.maxTime

            
            elif not isCurrentConFlag and isNextConFlag:
                if current_interval.minTime <= end <= next_interval.maxTime:
                    dur = min(end, next_interval.maxTime) - current_interval.minTime

                    if dur > best_dur:
                        best_dur = dur

                        left_boundary = current_interval.minTime
                        right_boundary = min(end, next_interval.maxTime)
            
        if left_boundary is None or right_boundary is None:
            left_boundary = current_interval.minTime
            right_boundary = next_interval.maxTime
        
        try:
            if intervals[idx+1].maxTime != intervals[idx+2].minTime and not isNextConFlag:
                right_boundary -= min_pause
                # print("触发最后一个intervalinterval")
        except IndexError:
            pass

        # valid_valleys = []
        # valley_valleys_rms = []
        mask = (valid_valleys >= left_boundary) & (valid_valleys <= right_boundary)

        while not valid_valleys[mask].any():
            right_boundary += 0.001
            left_boundary -= 0.001
            mask = (valid_valleys >= left_boundary) & (valid_valleys <= right_boundary)
        
        valid_valleys = valid_valleys[mask]
        valid_valleys_rms = valid_valleys_rms[mask]

        sorted_indices = np.argsort(valid_valleys_rms)
        
        if verbose:
            print(current_interval.mark, next_interval.mark)
            print(current_interval.minTime, next_interval.maxTime)
            print(left_boundary, right_boundary)


        if isNextConFlag or isCurrentConFlag:# or next_con in ["h", "d", "t", "k", "p", "f", "g", "n", "m", "b", "l"]:
            if next_con:
            # if isNextConFlag:
            #     valid_points = valid_valleys
            #     min_valley_time = valid_points[-1]
            # elif next_con and not isNextConFlag:
                try:
                    valid_points = [valid_valleys[sorted_indices[idx_v]] for idx_v in range(1)]#, key=lambda x: abs(x - midpoint))
                    if verbose:
                        print(valid_valleys[sorted_indices])
                        print(f"找到最小值{valid_points}")
                    # if valid_points:
                    #     valid_points = [valid_valleys[np.argmin(valid_valleys_rms)]]
                    # else:
                    if not valid_points:
                        valid_points = valid_valleys
                except IndexError:
                    valid_points = valid_valleys
                # if isCurrentConFlag and not isNextConFlag:
                #     min_valley_time = valid_points[0]
                if not isCurrentConFlag and isNextConFlag:
                    # min_valley_time = valid_valleys[-1]
                    min_valley_time = valid_points[0]

                else:
                    min_valley_time = valid_points[0]
            else:
                min_valley_time = find_internsity_valley(audio_path, left_boundary, right_boundary, verbose=verbose)


        else: # 两个都不是
            # if not next_con:
            #     min_valley_time = find_internsity_valley(audio_path, left_boundary, right_boundary, verbose=False)
            # else:
            min_valley_time = find_internsity_valley(audio_path, left_boundary, right_boundary, verbose=verbose)
        if verbose:
            print(min_valley_time)
            print()
        
        
        # elif next_con in ["k", "t", "p"]:
        #     min_valley_time = valid_valleys[np.argmin(valid_valleys_rms)]
        
        # elif not next_con or next_con in ["h"]:
        #     min_valley_time = find_internsity_valley(audio_path, current_interval.minTime, next_interval.maxTime)


        
        # print(f"最小波谷时间: {min_valley_time}")
        
        current_interval.maxTime = min_valley_time
        next_interval.minTime = current_interval.maxTime

        if verbose:
            # for v in valid_valleys:
            plt.axvline(x=min_valley_time, color='b', label="Valid" if idx == 0 else "", alpha=0.3, linewidth=2)
            # print()



    if verbose:
        # plt.legend()
        
        # 显示图形
        plt.tight_layout()
        plt.show()

        tg.write(tg_path.replace("_whisper.TextGrid", "_whisper_recali.TextGrid"))
    return tg

def get_baseline_freq_peak(audio_path):
    y, sr = read_audio(audio_path)
    # 计算每个帧的频谱
    stft = librosa.stft(y)#, win_length=win_length, hop_length=hop_length)
    # 计算每个帧的幅度谱
    magnitude = np.abs(stft)
    # 找到每个帧中频谱幅度最大的索引
    max_magnitude_indices = np.argmax(magnitude, axis=0)
    # 将索引转换为对应的频率
    spectral_peaks = librosa.fft_frequencies(sr=sr)[max_magnitude_indices]
    
    # 计算对应的时间轴
    time = librosa.times_like(spectral_peaks, sr=sr)#, hop_length=hop_length)
    
    
    vad_path = audio_path.replace(".wav", "_VAD.TextGrid")
    tg = TextGrid()
    tg.read(vad_path)
    
    vad_tiers = [(i.minTime, i.maxTime) for i in tg.tiers[0] if i.mark == ""]
    freq_peaks = []
    for start_time, end_time in vad_tiers:
        mask = (time >= start_time) & (time <= end_time)
        time = time[mask]
        spectral_peaks = spectral_peaks[mask]

        freq_peaks += list(spectral_peaks)
    return np.mean(freq_peaks) * 2.5
    
    


# 使用示例
if __name__ == "__main__":
    audio_file_path = r"C:\Users\User\Desktop\Praasper\data\mandarin_sent.wav" 
    # audio_file_path = r"C:\Users\User\Desktop\Praasper\data\test_audio.wav" 
    # audio_file_path = r"C:\Users\User\Desktop\Praasper\data\man_clip.wav" 

    # min_valley_time = find_internsity_valley(audio_file_path, 0, 8.0, verbose=True)
    # peak = find_spec_peak(audio_file_path, verbose=False)
    # get_baseline_freq_peak(audio_file_path)
    # exit()
    find_word_boundary(audio_file_path, tar_sr=12000, verbose=True)



try:
    from .utils import *
    from .process import *
    from .word_boundary import *
except ImportError:
    from utils import *
    from process import *
    from word_boundary import *

import os
import whisper
import torch

class init_model:

    def __init__(self, model_name: str="large-v3-turbo"):

        self.name = model_name

        available_models = whisper.available_models()
        if self.name in available_models:
            print(f"[{show_elapsed_time()}] Loading Whisper model: {self.name}")
        else:
            raise ValueError(f"[{show_elapsed_time()}] Model {self.name} is not in the available Whisper models. Available models are: {available_models}")
        
        device = "cuda" if torch.cuda.is_available() else "cpu"
        self.whisper_model = whisper.load_model(self.name, device=device)
        print(f"[{show_elapsed_time()}] Model loaded successfully. Current device in use: {self.whisper_model.device if hasattr(self.whisper_model, 'device') else 'Unknown'}")

    def annote(self, input_path: str, sr=None, language=None, verbose: bool=False):

        fnames = [os.path.splitext(f)[0] for f in os.listdir(input_path) if f.endswith('.wav')]
        print(f"[{show_elapsed_time()}] {len(fnames)} valid audio files detected in {input_path}")

        for idx, fname in enumerate(fnames):
            wav_path = os.path.join(input_path, fname + ".wav")
            tg_path = wav_path.replace(".wav", "_whisper.TextGrid")
            vad_path = wav_path.replace(".wav", "_VAD.TextGrid")

            output_path = os.path.join(os.path.dirname(os.path.dirname(tg_path)), "output")
            os.makedirs(output_path, exist_ok=True)
            final_path = os.path.join(output_path, os.path.basename(wav_path).replace(".wav", ".TextGrid"))

            final_tg = None
            print(f"--------------- Processing {os.path.basename(wav_path)} ({idx+1}/{len(fnames)}) ---------------")
            try:
                get_vad(wav_path, verbose=verbose)
            except Exception as e:
                print(f"[{show_elapsed_time()}] Error processing {os.path.basename(wav_path)}: {e}")
                continue
            try:
                language, final_tg = transcribe_wav_file(wav_path, vad=vad_path, whisper_model=self.whisper_model, language=language)
            except Exception as e:
                print(f"[{show_elapsed_time()}] Error transcribing {os.path.basename(wav_path)}: {e}")
                continue
            try:
                final_tg = find_word_boundary(wav_path, tar_sr=sr, verbose=verbose)
            except Exception as e:
                print(f"[{show_elapsed_time()}] Error finding word boundary {os.path.basename(wav_path)}: {e}")
            if final_tg:
                final_tg.write(final_path)
        
        print(f"--------------- Processing completed ---------------")


if __name__ == "__main__":
    model = init_model(model_name="large-v3-turbo")
    model.annote(input_path=os.path.abspath("data"), sr=12000, language=None, verbose=False)


import os
from textgrid import TextGrid, IntervalTier
import librosa
import numpy as np
from scipy.signal import convolve2d, find_peaks, butter, filtfilt
import torch

try:
    import matplotlib.pyplot as plt
except ImportError:
    pass

try:
    from .utils import *
    from .VAD.core_auto import *
# from .praditor.tool_auto import * 
except ImportError:
    from utils import *
    from VAD.core_auto import *
from scipy.signal import argrelextrema
import os




def detect_energy_valleys(wav_path, tg_path):
    try:
        y, sr = librosa.load(wav_path, mono=False, sr=16000)
        y = y[0]
        y = np.gradient(np.gradient(y))
    except ValueError:
        y, sr = librosa.load(wav_path, mono=True, sr=16000)
        y = np.gradient(np.gradient(y))

    

    # 使用librosa计算音频的均方根能量(rms)
    rms = librosa.feature.rms(y=y, frame_length=128, hop_length=32, center=True)[0]
    
    # 计算对应的时间轴
    time = librosa.times_like(rms, sr=sr, hop_length=32)

    
    # 找到波谷的索引
    valley_indices = find_peaks(-rms, width=(5, None))[0]


    tg_vad = TextGrid()
    tg_vad.read(tg_path)
    intervals = [interval for interval in tg_vad.tiers[0] if interval.mark != ""]


    # 定义筛选规则：
    # 1. 波谷中的波谷/拐点
    # 2. 不允许左高右低
    for idx, interval in enumerate(intervals):
        if idx == len(intervals) - 1:
            break
        
        current_interval = intervals[idx]
        next_interval = intervals[idx+1]

        if current_interval.maxTime != next_interval.minTime:
            continue

        cand_valleys = [t for t in time[valley_indices] if current_interval.minTime + 0.01 < t < next_interval.maxTime - 0.01]

        # 获取 cand_valleys 对应的 rms 值
        cand_valleys_rms = [rms[np.where(time == t)[0][0]] for t in cand_valleys]
        print()
        print(cand_valleys, cand_valleys_rms)

        # 筛选出左相邻小于右相邻的波谷
        valid_valleys = []
        valid_valleys_rms = []
        if len(cand_valleys) >= 3:

            for idx_in_time, t in enumerate(cand_valleys):
                if idx_in_time == len(cand_valleys) - 1:
                    continue
                else:

                    if cand_valleys_rms[idx_in_time] < cand_valleys_rms[idx_in_time+1]:
                        valid_valleys.append(t)
                        valid_valleys_rms.append(rms[idx_in_time])
            
            if not valid_valleys:
                valid_valleys = cand_valleys
                valid_valleys_rms = cand_valleys_rms

        else:
            valid_valleys = cand_valleys
            valid_valleys_rms = cand_valleys_rms




        min_valley_time = valid_valleys[np.argmin(valid_valleys_rms)]
        
        current_interval.maxTime = min_valley_time
        next_interval.minTime = current_interval.maxTime
             
        # print(valid_valleys, valid_valleys_rms)
    tg_vad.write(tg_path.replace(".TextGrid", "_recali.TextGrid"))


def calc_power_valley(y, sr):
    """
    用librosa库计算一段信号的功率曲线，并找到最低的波谷所在的时间戳

    :param y: 音频信号
    :param sr: 采样率
    :return: 功率曲线, 最低波谷对应的时间戳
    """

    # 使用librosa计算音频的均方根能量(rms)
    rms = librosa.feature.rms(y=y, frame_length=512, hop_length=256, center=False)[0]
    
    # 计算对应的时间轴
    time = librosa.times_like(rms, sr=sr, hop_length=256)
    # 找到所有波谷的索引
    valley_indices = argrelextrema(rms, np.less)[0]
        
    # 获取所有波谷对应的时间戳
    valley_time = time[valley_indices]
    
    
    return rms, valley_time



def bandpass_filter(data, lowcut, highcut, fs, order=4):
    nyquist = 0.5 * fs
    low = lowcut / nyquist
    high = highcut / nyquist
    if low == 0:
        b, a = butter(order, high, btype='low', output="ba")
        filtered_data = filtfilt(b, a, data)
    else:
        try:
            b, a = butter(order, [low, high], btype='bandpass', output="ba")
            filtered_data = filtfilt(b, a, data)
        except ValueError:  # 如果设置的最高频率大于了可接受的范围
            b, a = butter(order, low, btype='high', output="ba")
            filtered_data = filtfilt(b, a, data)
    return filtered_data


def find_valid_peaks(y, sr, db_thre=20, peak_prominence=10, nfft=2048, win_length=1024):
    # 计算频谱图
    spectrogram = librosa.stft(y, n_fft=nfft, win_length=win_length, center=True)
    spectrogram_db = librosa.amplitude_to_db(abs(spectrogram), ref=1.0)  # 使用librosa.amplitude_to_db已将y值转换为对数刻度，top_db=None确保不限制最大分贝值
    
    kernel = np.array([[-1, 0, 1]])
    convolved_spectrogram = convolve2d(spectrogram_db, kernel, mode='same', boundary='symm')
    convolved_spectrogram = np.where(np.abs(convolved_spectrogram) < db_thre, 0, convolved_spectrogram)

    # 按频率轴求和，保持维度以方便后续绘图
    convolved_spectrogram = np.sum(np.abs(convolved_spectrogram), axis=0, keepdims=False)
    # 在保持输出信号长度不变的情况下，对卷积后的频谱图求一阶导
    # convolved_spectrogram = np.gradient(convolved_spectrogram)
    time_axis = np.linspace(0, len(convolved_spectrogram) * librosa.core.get_duration(y=y, sr=sr) / len(convolved_spectrogram), len(convolved_spectrogram))

    # 找到所有的波峰和波谷
    peaks, _ = find_peaks(convolved_spectrogram, prominence=(peak_prominence, None))
    # valleys, _ = find_peaks(-convolved_spectrogram, prominence=(10, None))

    # valid_valleys = valleys[np.abs(convolved_spectrogram[valleys]) > 0]

    # 提取有效波峰和波谷对应的时间和值
    # peak_times = time_axis[valid_peaks]

    return peaks, time_axis, convolved_spectrogram


def compare_centroids(y, time_stamp, sr, length=0.05):
    """
    比较前后的音频片段的频谱能量质心
    """

    target_sample = int(time_stamp * sr)
    start_prev = target_sample - int(length * sr)
    end_prev = target_sample
    start_next = target_sample
    end_next = target_sample + int(length * sr)

    # 提取前后0.01s的音频片段
    y_prev = y[start_prev:end_prev]
    y_next = y[start_next:end_next]
    
    # 计算前0.01s的频谱能量质心
    centroid_prev = librosa.feature.spectral_centroid(y=y_prev, sr=sr)
    centroid_prev_mean = np.mean(centroid_prev)
    
    # 计算后0.01s的频谱能量质心
    centroid_next = librosa.feature.spectral_centroid(y=y_next, sr=sr)
    centroid_next_mean = np.mean(centroid_next)


    # 计算前0.01s的音频能量总量
    energy_prev = np.sum(y_prev ** 2)
    
    # 计算后0.01s的音频能量总量
    energy_next = np.sum(y_next ** 2)

    return centroid_prev_mean, centroid_next_mean, energy_prev, energy_next


def get_vad(wav_path, min_pause=0.2, params="self", verbose=False):

    print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD processing started...")


    audio_obj = ReadSound(wav_path)

    # 获取 wav 文件所在的文件夹路径
    wav_folder = os.path.dirname(wav_path)
    all_txt_path = os.path.join(wav_folder, "params.txt")
    self_txt_path = wav_path.replace(".wav", ".txt")

    default_params = {'onset': {'amp': '1.47', 'cutoff0': '60', 'cutoff1': '10800', 'numValid': '475', 'eps_ratio': '0.093'}, 'offset': {'amp': '1.47', 'cutoff0': '60', 'cutoff1': '10800', 'numValid': '475', 'eps_ratio': '0.093'}}
    

    if params == "all":
        if os.path.exists(all_txt_path):
            with open(all_txt_path, "r") as f:
                params = eval(f.read())
        else:
            params = default_params
    
    elif params == "self":
        if os.path.exists(self_txt_path):
            with open(self_txt_path, "r") as f:
                params = eval(f.read())
        else:
            params = default_params
    elif params == "default":
        params = default_params

    else:  # 具体参数
        params = params
    



    onsets = autoPraditorWithTimeRange(params, audio_obj, "onset")
    offsets = autoPraditorWithTimeRange(params, audio_obj, "offset")
    if verbose:
        print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD onsets: {onsets}")
        print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD offsets: {offsets}")

    valid_onsets = onsets[:1]
    valid_offsets = []
    for idx, xset in enumerate(onsets):
        if idx == len(onsets) - 1:
            break
        onset = onsets[idx+1]
        offset = offsets[idx]

        if onset - offset > min_pause:
            valid_onsets.append(onset)
            valid_offsets.append(offset)
    valid_offsets.append(offsets[-1])

    onsets = valid_onsets
    offsets = valid_offsets

    tg = TextGrid()
    interval_tier = IntervalTier(name="interval", minTime=0., maxTime=audio_obj.duration_seconds)
    for i in range(len(onsets)):
        try:
            interval_tier.add(onsets[i], offsets[i], "+")
        except ValueError:
            continue
        # except IndexError:
        #     break
    tg.append(interval_tier)
    tg.write(wav_path.replace(".wav", "_VAD.TextGrid"))  # 将TextGrid对象写入文件
    print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD results saved")

# else:


# defs
def transcribe_wav_file(wav_path, vad, whisper_model, language):
    """
    使用 Whisper 模型转录 .wav 文件
    
    :param file_path: .wav 文件的路径
    :param path_vad: VAD TextGrid 文件的路径
    :return: 转录结果
    """

    # 转录音频文件
    initial_prompt = """
    请保留所有语气词，比如嗯、啊、呃、唉、呵、呼、哼、咳、呜、哇、呀、喔、哦、哎、嘛。
    """
    if language != None:
        result = whisper_model.transcribe(wav_path, initial_prompt=initial_prompt, fp16=torch.cuda.is_available(), word_timestamps=True, language=language)
    else:
        result = whisper_model.transcribe(wav_path, initial_prompt=initial_prompt, fp16=torch.cuda.is_available(), word_timestamps=True)
    
    language = result["language"]
    print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) Transcribing into {language}...")
    # print(result)

    # 加载 path_vad 对应的 TextGrid 文件
    try:
        vad_tg = TextGrid.fromFile(vad)
    except FileNotFoundError:
        print(f"错误：未找到文件 {vad}")
        raise

    # 提取所有 mark 为空字符串的 interval 的起止时间
    vad_intervals = []
    empty_mark_intervals = []
    for tier in vad_tg:
        for interval in tier:
            if interval.mark == "":
                empty_mark_intervals.append((interval.minTime, interval.maxTime))
            else:
                vad_intervals.append((interval.minTime, interval.maxTime))



    tg = TextGrid()
    tier = IntervalTier(name='word', minTime=0.0, maxTime=vad_tg.tiers[0].maxTime)

    for segment in result["segments"]:
        
        for idx, word in enumerate(segment["words"]):
            start_time = word["start"]
            end_time = word["end"]
            
            text = ''.join(c for c in word["word"] if c.isalpha() or c.isspace() or c.isnumeric())  # 去掉标点符号

            # print(start_time, end_time, text)
            for empty_mark_interval in empty_mark_intervals:
                # print(empty_mark_interval)
                if empty_mark_interval[0] <= end_time <= empty_mark_interval[1]:
                    end_time = empty_mark_interval[0]
                
                if empty_mark_interval[0] <= start_time <= empty_mark_interval[1]:
                    start_time = empty_mark_interval[1]
                
                if start_time < empty_mark_interval[0] < empty_mark_interval[1] < end_time:
                    pass

            tier.add(start_time, end_time, text)

    for vad_interval in vad_intervals:
        # 找到距离 vad_interval[0] 最近的 interval.minTime
        closest_interval = min(tier.intervals, key=lambda x: abs(x.minTime - vad_interval[0]))

        if closest_interval.minTime - vad_interval[0] != 0:
            closest_interval.minTime = vad_interval[0]

        # 找到距离 vad_interval[1] 最近的 interval.maxTime
        closest_interval = min(tier.intervals, key=lambda x: abs(x.maxTime - vad_interval[1]))

        if closest_interval.maxTime - vad_interval[1] != 0:
            closest_interval.maxTime = vad_interval[1]

    # 检查tier里是否有mark=”+“的interval，若有则删除
    tier.intervals = [interval for interval in tier.intervals if interval.mark != "+"]
    # print(tier.intervals)
    # 遍历 word_tier 中的每一个 interval
    new_intervals = []
    for interval in tier.intervals:
        if len(interval.mark) > 1:
            # 计算每个新 interval 的时长
            duration = (interval.maxTime - interval.minTime) / len(interval.mark)
            start_time = interval.minTime
            # 将每个中文字拆分为一个新的 interval
            for char in interval.mark:
                end_time = start_time + duration
                new_interval = Interval(start_time, end_time, char)
                new_intervals.append(new_interval)
                start_time = end_time
        else:
            new_intervals.append(interval)
    # 替换原有的 intervals
    tier.intervals = new_intervals


    for idx, interval in enumerate(tier.intervals):
        if idx == len(tier.intervals) -1:
            break


        current_interval = tier.intervals[idx]
        next_interval = tier.intervals[idx + 1]

        if next_interval.minTime < current_interval.maxTime and next_interval.mark != "":
            current_interval.maxTime = next_interval.minTime

    tg.append(tier)
    tg.write(wav_path.replace(".wav", "_whisper.TextGrid"))
    print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) Whisper word-level transcription saved")
    return language, tg


def word_timestamp(wav_path, tg_path, language):

    if language.lower() not in ['zh', 'en', 'yue']:
        print(f"[{show_elapsed_time()}] Language {language} not currently supported.")

        wav_folder = os.path.dirname(os.path.dirname(wav_path))
        output_path = os.path.join(wav_folder, "output")
        os.makedirs(output_path, exist_ok=True)
        new_tg_path = os.path.join(output_path, os.path.basename(wav_path).replace(".wav", ".TextGrid"))
        tg.write(new_tg_path)
        return


    print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) Trimming word-level annotation...")
    detect_energy_valleys(wav_path, tg_path)

    return
    y, sr = librosa.load(wav_path, sr=16000)  # 加载音频文件
    # y = bandpass_filter(y, 200, 8000, sr)
    
    tg = TextGrid.fromFile(tg_path)  # 加载 TextGrid 文件

    word_tier = [tier for tier in tg if tier.name == 'word'][0]

    # 计算 tg 的 segment 中 mark 不为空的 interval 的平均时长
    non_empty_intervals = [interval.maxTime - interval.minTime for tier in tg for interval in tier if interval.mark != ""]
    average_word_duration = np.mean(non_empty_intervals) if non_empty_intervals else 0
    
    word_intervals = [interval for interval in word_tier.intervals if interval.mark != ""]
    for i in range(len(word_intervals) - 1):
        current_interval = word_intervals[i]
        next_interval = word_intervals[i + 1]
        # 检查两个 interval 是否相粘着（前一个的结束时间等于后一个的开始时间）

        if current_interval.maxTime == next_interval.minTime:
            target_boundary = current_interval.maxTime - current_interval.minTime

            start_sample = int(current_interval.minTime * sr)
            end_sample = int(next_interval.maxTime * sr)
            y_vad = y[start_sample:end_sample]

            peaks, time_axis, convolved_spectrogram = find_valid_peaks(y_vad, sr)

            peak_times = time_axis[peaks]

            # 筛选出不在 current_interval.minTime 到 current_interval.minTime + 0.05s 之间的波峰
            valid_peak_times = [t for t in peak_times if t >= 0.05 and (target_boundary -  average_word_duration/2 <= t <= target_boundary + average_word_duration * 3/4)]

            if valid_peak_times:
                # 找到距离 target_boundary 最近且最大的波峰
                # 获取波峰对应的数值
                peak_values_nearby = [convolved_spectrogram[int((t / librosa.core.get_duration(y=y_vad, sr=sr)) * len(convolved_spectrogram))] for t in valid_peak_times]
                # 找到最大波峰对应的时间
                closest_peak_time = valid_peak_times[np.argmax(peak_values_nearby)]
            else:
                closest_peak_time = target_boundary
            
            # 找到之后，开始写入
            target_boundary = closest_peak_time + current_interval.minTime

            current_interval.maxTime = target_boundary
            next_interval.minTime = target_boundary
            print(target_boundary)
            print("SR:", sr)
            # 以target_boundary为0点，分别计算-0.01s到0s以及0s到0.01s的频谱能量质心
            
            centroid_prev_mean, centroid_next_mean, energy_prev, energy_next = compare_centroids(y, target_boundary, sr)

            print(current_interval.mark, next_interval.mark, f"{centroid_prev_mean:.2f} Hz", f"{centroid_next_mean:.2f} Hz", f"{energy_prev:.2f}", f"{energy_next:.2f}")

            if centroid_prev_mean > centroid_next_mean and energy_prev < energy_next:
                next_cvt = extract_cvt(next_interval.mark, lang=language)[0]
                print(next_interval.mark, next_cvt)
                if next_cvt[0] in ["d", "l"]:
                    y_alt = bandpass_filter(y, 4000, 6000, sr)
                elif not next_cvt[0]:
                    y_alt = bandpass_filter(y, 0, 2000, sr)
                else:
                    y_alt = y


                # 从 target_boundary 往前，在 peak_times 里找第一个左质心小于右质心的峰值
                # 将峰值时间按从大到小排序，从靠近 target_boundary 的峰值开始检查
                sorted_peak_times = sorted([t for t in peak_times if target_boundary - current_interval.minTime < t < next_interval.maxTime - current_interval.minTime]) + \
                                    sorted([t for t in peak_times if 0.05 < t < target_boundary - current_interval.minTime - 0.000001], reverse=True) # 要做两个方向
                print(sorted_peak_times)
# !

                for peak_time in sorted_peak_times:
                    # 计算该峰值对应的实际时间点
                    actual_peak_time = peak_time + current_interval.minTime
                    
                    new_centroid_prev_mean, new_centroid_next_mean, new_energy_prev, new_energy_next = compare_centroids(y_alt, actual_peak_time, sr)
                    print("\t", actual_peak_time, f"{new_centroid_prev_mean:.2f} Hz", f"{new_centroid_next_mean:.2f} Hz", f"{new_energy_prev:.2f}", f"{new_energy_next:.2f}")
                    if next_cvt[0] in ["d", "l"]:
                        if new_energy_prev < new_energy_next:
                            target_boundary = actual_peak_time
                            # print("调整后", target_boundary, f"{centroid_prev_mean} Hz", f"{centroid_next_mean} Hz")
                            current_interval.maxTime = target_boundary
                            next_interval.minTime = target_boundary
                            break
                    
                    else:
                        if new_centroid_prev_mean < new_centroid_next_mean: # and new_energy_prev < new_energy_next:
                            target_boundary = actual_peak_time
                            # print("调整后", target_boundary, f"{centroid_prev_mean} Hz", f"{centroid_next_mean} Hz")
                            current_interval.maxTime = target_boundary
                            next_interval.minTime = target_boundary
                            break
                    
            print()

    # y, sr = librosa.load(wav_path, sr=18000)


    tg.write(wav_path.replace(".wav", "_recali.TextGrid"))

    phon_tier = IntervalTier(name="phoneme", minTime=0, maxTime=word_tier.maxTime)

    for interval in word_intervals:

        con, vow, tone = extract_cvt(interval.mark, lang=language)[0]
        # !先尝试分开声韵母

        # print(cvt)
        start_sample = int(interval.minTime * sr)
        end_sample = int(interval.maxTime * sr)
        # print(interval.mark, interval.minTime, interval.maxTime)

        y_vad = y[start_sample:end_sample]

        expected_num = 2 if con else 1 #len(vow) + 1 if con else len(vow)
        # print(expected_num)

        peaks, time_axis, convolved_spectrogram = find_valid_peaks(y_vad, sr)

        # 按峰值大小对峰值索引进行排序
        # 筛选出左能量质心高于右能量质心的 peak
        valid_peaks = []
        for peak in peaks:
            peak_time = time_axis[peak] + interval.minTime
            centroid_prev_mean, centroid_next_mean, energy_prev, energy_next = compare_centroids(y, peak_time, sr)
            # print(f"频谱能量质心（-0.01s到0s）: {centroid_prev_mean} Hz | 频谱能量质心（0s到0.01s）: {centroid_next_mean} Hz | 频谱能量（-0.01s到0s）: {energy_prev} | 频谱能量（0s到0.01s）: {energy_next}")
            if centroid_prev_mean > centroid_next_mean: # and energy_prev > energy_next:
                valid_peaks.append(peak)
        
        # 获取前 expected_num-1 个最大的符合条件的 peak
        sorted_peaks = sorted(valid_peaks, key=lambda x: convolved_spectrogram[x], reverse=True)[:expected_num-1]


        # 获取波峰对应的时间戳
        peak_times = time_axis[sorted_peaks]

        peak_timestamps = [interval.minTime] + [pt + interval.minTime for pt in peak_times] + [interval.maxTime]

        # peak_timestamps.sort()
        peak_timestamps = sorted(peak_timestamps, reverse=True)
        # print(peak_timestamps)

        for t, time_stamp in enumerate(peak_timestamps):
            if t == 0:
                phon_tier.add(peak_timestamps[t+1], peak_timestamps[t], "".join(vow))
            else:
                try:
                    phon_tier.add(peak_timestamps[t+1], peak_timestamps[t], con)
                except IndexError:
                    pass
    # print(tg.maxTime)
    tg.append(phon_tier)


    # 保存修改后的 TextGrid 文件
    # 检查 output 文件夹是否存在，如果不存在则创建
    wav_folder = os.path.dirname(os.path.dirname(wav_path))
    output_path = os.path.join(wav_folder, "output")
    os.makedirs(output_path, exist_ok=True)
    new_tg_path = os.path.join(output_path, os.path.basename(wav_path).replace(".wav", ".TextGrid"))
    tg.write(new_tg_path)
    print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) Phoneme-level segmentation saved")

    return tg






import pypinyin
try:
    from pycantonese import characters_to_jyutping
except ImportError:
    pass


def extract_cvt_yue(character):
    """
    给定一个中文单字，返回其对应的粤语拼音（声母、韵母、声调）
    
    :param character: 单个中文字符
    :return: 包含声母、韵母、声调的字典
    """
    if len(character) != 1:
        return ("", [character], "")
    
    # 导入放在函数内避免依赖问题（实际使用时建议放在顶部）
    
    # 获取粤语拼音（Jyutping）
    jyutping_list = characters_to_jyutping(character)
    if not jyutping_list:
        return ("", [], "")
    
    # 取第一个读音（忽略多音字）
    full_pinyin = jyutping_list[0][0]
    
    # 分离声调（Jyutping 声调在末尾数字1-6）
    tone = ""
    for char in reversed(full_pinyin):
        if char.isdigit():
            tone = char
            base = full_pinyin.rstrip('123456')
            break
    else:
        base = full_pinyin
    
    # 分离声母和韵母
    initials = ["b", "p", "m", "f", "d", "t", "n", "l", "g", "k", "ng", "h", 
                "gw", "kw", "w", "z", "c", "s", "j"]
    initial = ""
    final = base
    
    # 检查最长匹配（先匹配2字母声母）
    for s in sorted(initials, key=len, reverse=True):
        if base.startswith(s):
            initial = s
            final = base[len(s):]
            break
    
    # 处理韵母（将字符串转为列表并合并 ng）
    final_list = []
    i = 0
    while i < len(final):
        if i < len(final) - 1 and final[i:i+2] == "ng":
            final_list.append("ng")
            i += 2
        else:
            final_list.append(final[i])
            i += 1
    
    return (
        initial,
        final_list,
        tone
    )


def extract_cvt_zh(character):
    """
    给定一个中文单字，返回其对应的拼音（声母、韵母、声调）
    
    :param character: 单个中文字符
    :return: 包含声母、韵母、声调的字典
    """
    # if len(character) != 1:
        # raise ValueError("只能输入单个中文字符")
    cvts = []
    for char in character:
        # 获取拼音信息
        pinyin_result = pypinyin.pinyin(char, style=pypinyin.TONE3, heteronym=False)[0][0]
        
        # 提取声母、韵母和声调
        initial = pypinyin.pinyin(char, style=pypinyin.INITIALS, heteronym=False)[0][0]
        final_with_tone = pypinyin.pinyin(char, style=pypinyin.FINALS_TONE3, heteronym=False)[0][0]
        
        # 分离韵母和声调
        tone = ''
        final = final_with_tone
        for char in final_with_tone:
            if char.isdigit():
                tone = char
                final = final_with_tone.replace(char, '')
                break
        
        

        final = list(final)
        # 如果n和g相邻，合并成ng
        new_final = []
        i = 0
        while i < len(final):
            if i < len(final) - 1 and final[i] == 'n' and final[i+1] == 'g':
                new_final.append('ng')
                i += 2
            else:
                new_final.append(final[i])
                i += 1
        final = new_final

        cvts.append((
            initial if initial else '',
            final,
            tone if tone else ''
        ))
    
    return cvts

if __name__ == "__main__":
    print(extract_cvt_zh('我'))


import os.path
import sys
from datetime import datetime

import numpy as np
from scipy.signal import butter, filtfilt
from textgrid import TextGrid
# import soundfile as sf
import librosa


class ReadSound:
    def __init__(self, fpath=None, arr=None, duration_seconds=None, frame_rate=None):

        self.fpath = fpath
        if fpath is None:
            if arr is None:
                raise ValueError("Need audio input. Receive None.")
            else:
                self.arr = arr
                self.duration_seconds = duration_seconds
                self.frame_rate = frame_rate

        else:  # 如果有fpath
            self.arr, self.frame_rate = librosa.load(fpath, sr=None, dtype=np.float32)
            self.arr = (self.arr * 32767).astype('int16')  # 转换为 int16 类型
            self.duration_seconds = librosa.get_duration(y=self.arr, sr=self.frame_rate)



        try:
            self.arr = self.arr[:, 0]
        except IndexError:
            pass

        self.max = np.max(np.abs(self.arr))

    def __getitem__(self, ms):


        start = int(ms.start * self.frame_rate / 1000) if ms.start is not None else 0
        end = int(ms.stop * self.frame_rate / 1000) if ms.stop is not None else len(self.arr)

        start = min(start, len(self.arr))
        end = min(end, len(self.arr))


        return ReadSound(arr=self.arr[start:end], duration_seconds=(end - start) / 1000, frame_rate=self.frame_rate)

    def get_array_of_samples(self):
        return self.arr



def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")
    return os.path.join(base_path, relative_path)


def get_current_time():
    # 获取当前时间
    now = datetime.now()
    # 格式化时间字符串
    formatted_time = now.strftime("%H:%M:%S.%f")[:-3]  # 去掉微秒部分的最后3个字符
    return formatted_time


def bandpass_filter(data, lowcut, highcut, fs, order=4):
    nyquist = 0.5 * fs
    low = lowcut / nyquist
    high = highcut / nyquist
    if low == 0:
        b, a = butter(order, high, btype='low', output="ba")
        filtered_data = filtfilt(b, a, data)
    else:
        try:
            b, a = butter(order, [low, high], btype='bandpass', output="ba")
            filtered_data = filtfilt(b, a, data)
        except ValueError:  # 如果设置的最高频率大于了可接受的范围
            b, a = butter(order, low, btype='high', output="ba")
            filtered_data = filtfilt(b, a, data)
    return filtered_data


def lowpass_filter(data, highcut, fs, order=4):
    nyquist = 0.5 * fs
    high = highcut / nyquist
    b, a = butter(order, high, btype='low', output="ba")
    filtered_data = filtfilt(b, a, data)
    return filtered_data


def isAudioFile(fpath):
    # 所有的音频后缀
    audio_extensions = [
        '.mp3',  # MPEG Audio Layer-3
        '.wav',   # Waveform Audio File Format
        ".WAV",
        '.ogg',   # Ogg
        '.flac',  # Free Lossless Audio Codec
        '.aac',   # Advanced Audio Codec
        '.m4a',   # MPEG-4 Audio Layer
        '.alac',  # Apple Lossless Audio Codec
        '.aiff',  # Audio Interchange File Format
        '.au',    # Sun/NeXT Audio File Format
        '.aup',   # Audio Unix/NeXT
        '.ra',    # RealAudio
        '.ram',   # RealAudio Metafile
        '.rv64',  # Raw 64-bit float (AIFF/AIFF-C)
        '.spx',   # Ogg Speex
        '.voc',   # Creative Voice
        '.webm',  # WebM (audio part)
        '.wma',   # Windows Media Audio
        '.xm',    # FastTracker 2 audio module
        '.it',    # Impulse Tracker audio module
        '.mod',   # Amiga module (MOD)
        '.s3m',   # Scream Tracker 3 audio module
        '.mtm',   # MultiTracker audio module
        '.umx',   # FastTracker 2 extended module
        '.dxm',   # Digital Tracker (DTMF) audio module
        '.f4a',   # FAudio (FMOD audio format)
        '.opus',  # Opus Interactive Audio Codec
    ]
    if any(fpath.endswith(ext) for ext in audio_extensions):
        return True
    else:
        return False


def get_frm_points_from_textgrid(audio_file_path):

    audio_dir = os.path.dirname(os.path.abspath(audio_file_path))
    audio_filename = os.path.splitext(os.path.basename(audio_file_path))[0]
    tg_file_path = os.path.join(audio_dir, audio_filename + ".TextGrid")
    if not os.path.exists(tg_file_path):
        return {"onset":[], "offset": []}
    tg = TextGrid(tg_file_path)
    tg.read(tg_file_path)
    dict_tg_time = {}
    for tier in tg.tiers:
        if tier.name == "interval":
            continue
        dict_tg_time[tier.name] = [p.time for p in tier]
    return dict_tg_time


def get_frm_points_from_01(file_path):
    if not os.path.exists(file_path):
        return None
    tg = TextGrid(file_path)
    tg.read(file_path)
    dict_tg = {}

    for tier in tg.tiers:
        dict_tg["onset"] = [p.minTime for p in tier if p.mark == "1"]
        dict_tg["offset"] = [p.maxTime for p in tier if p.mark == "1"]

    return dict_tg


def get_interval(file_path):
    tg = TextGrid(file_path)
    tg.read(file_path)

    tg_onsets = []
    tg_offsets = []

    for tier in tg.tiers:
        if tier.name == "onset":
            tg_onsets = [p.time * 1000 for p in tier]
        else:
            tg_offsets = [p.time * 1000 for p in tier]

    return list(zip(tg_onsets, tg_offsets))


def get_time(file_path):
    # 加载TextGrid文件
    tg = TextGrid(file_path)
    tg.read(file_path)
    # 找到名为'onset'的PointTier
    point_tier_onset = None

    for tier in tg.tiers:
        if tier.name == 'onset':
            point_tier_onset = tier
            break

    # 如果找到了PointTier，获取第一个点的时间
    if point_tier_onset:
        return point_tier_onset.points[0].time
        # print(f"The time of the first point in 'onset' tier is: {first_point_time}")
    else:
        return None


try:
    from .utils import *
    from .cvt import *
except ImportError:
    from utils import *
    from cvt import *


from numba.core.compiler import PassManager
from textgrid import TextGrid, IntervalTier

def find_phon_peak(audio_path, tg_path, lang, verbose=False):
    tg = TextGrid()
    tg.read(tg_path)
    intervals = [interval for interval in tg.tiers[0] if interval.mark != ""]

    phon_tier = IntervalTier(name="phon", maxTime=tg.tiers[0].maxTime)

    for idx, interval in enumerate(intervals):
        con, vow, tone = extract_cvt(interval.mark, lang=lang)[0]

        if con == "":
            phon_tier.addInterval(Interval(interval.minTime, interval.maxTime, vow + tone))
        elif con in ["b", "t", "p", "k"]:
            pass

        elif con in ["z", "zh", "s", "sh", "c", "ch", "j", "x"]:
            pass
            


        



from setuptools import setup, find_packages

setup(
    name='praasper',      # 在PyPI上唯一的包名
    version='0.2.2',              # 版本号，遵循语义化版本规范
    description='VAD-Enhanced ASR with Word-Level Timestamps',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Tony Liu',
    author_email='paradeluxe3726@gmail.com',
    url='https://github.com/ParadeLuxe/Praasper',
    packages=find_packages(),     # 自动查找包
    install_requires=[            # 你的项目所依赖的第三方库
        'textgrid',
        'pypinyin',
        'librosa',
        'openai-whisper',
        'praat-parselmouth',
        'triton-windows; platform_system == "Windows"'
    ],
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',      # 指定Python版本要求
)